import team
import pandas as pd
import numpy as np
import tensorflow as tf 
import sys

physical_devices = tf.config.experimental.list_physical_devices('GPU')
if len(physical_devices) > 0:
  tf.config.experimental.set_memory_growth(physical_devices[0], True)


class Model(object):
    def __init__(self, model_dir, name, desc=''):
        self.model_dir = model_dir
        self.name = name
        self.desc = desc
        self.model = None
        self._load_model()
    
    def _load_model(self):
        try:
            self.model = tf.keras.models.load_model(self.model_dir)
        except:
            raise RuntimeError('Could not load model from {}'.format(self.model_dir))
    
    def get_model(self):
        return self.model


class MatchSimulator(object):
    def __init__(self, team1, team2, log_verbosity, model):
        self.bat_team = team1
        self.bowl_team = team2
        self.log_verbosity = log_verbosity
        self.model = model
        self.possible_outcomes = ['0','1','2','3','4','6','W']
        self._init_state()
    

    def _init_state(self):
        "Initializing several match state parameters"
        self.overs = 0
        self.balls = 0
        self.innings = 1
        #self.win_team = None
    

    def simulate_ball(self):
        #Simulates the outcome of a ball
        bowler = self.bowl_team.bowler
        batsman = self.bat_team.onstrike

        bowler_stats = self.bowl_team.get_hist_probs(bowler, 'bowl')
        batsman_stats = self.bat_team.get_hist_probs(batsman, 'bat')

        if self.innings == 1:
            target = 0
        else:
            target = self.bowl_team.bat_total

        model_inp = {
            'over':self.overs,
            'innings':self.innings,
            'score':self.bat_team.bat_total,
            'wickets_lost':self.bat_team.bat_wickets,
            'target':target,
            'bat_class_0':batsman_stats[0],
            'bat_class_1':batsman_stats[1],
            'bat_class_2':batsman_stats[2],
            'bat_class_3':batsman_stats[3],
            'bat_class_4':batsman_stats[4],
            'bat_class_5':batsman_stats[5],
            'bat_class_6':batsman_stats[6],
            'bowl_class_0':bowler_stats[0],
            'bowl_class_1':bowler_stats[1],
            'bowl_class_2':bowler_stats[2],
            'bowl_class_3':bowler_stats[3],
            'bowl_class_4':bowler_stats[4],
            'bowl_class_5':bowler_stats[5],
            'bowl_class_6':bowler_stats[6]         }
        
        input_dict = {name: tf.convert_to_tensor([value]) for name, value in model_inp.items()}
        predictions = self.model.predict(input_dict)
        output_probs = tf.nn.softmax(predictions[0])
        
        #sample the output probability distribution
        res = np.random.choice(a=self.possible_outcomes, size=1, p=np.array(output_probs))[0]
        #Apply updates as per the ball result
        self.update_match_state(res)
        
        return res
    


    def update_match_state(self, res):
        self.balls += 1
        
        if res == '0':
            if self.log_verbosity:
                print('Dot ball from {} to {}'.format(self.bowl_team.bowler, self.bat_team.onstrike))
        
        if res == '1':
            self.bat_team.player_match_stats[self.bat_team.onstrike]['runs_scored'] += 1
            self.bowl_team.player_match_stats[self.bowl_team.bowler]['runs_off'] += 1
            self.bat_team.bat_total += 1
            self.bowl_team.bowl_total += 1
            if self.log_verbosity:
                print('Single by {}. Score of {} is {}/{}'.format(self.bat_team.onstrike, self.bat_team.name, self.bat_team.bat_total, self.bat_team.bat_wickets))
            self.bat_team.change_strike()
        
        if res == '2':
            self.bat_team.player_match_stats[self.bat_team.onstrike]['runs_scored'] += 2
            self.bowl_team.player_match_stats[self.bowl_team.bowler]['runs_off'] += 2
            self.bat_team.bat_total += 2
            self.bowl_team.bowl_total += 2
            if self.log_verbosity:
                print('Two runs scored by {}. Score of {} is {}/{}'.format(self.bat_team.onstrike, self.bat_team.name, self.bat_team.bat_total, self.bat_team.bat_wickets))
        
        if res == '3':
            self.bat_team.player_match_stats[self.bat_team.onstrike]['runs_scored'] += 3
            self.bowl_team.player_match_stats[self.bowl_team.bowler]['runs_off'] += 3
            self.bat_team.bat_total += 3
            self.bowl_team.bowl_total += 3
            if self.log_verbosity:
                print('Three runs scored by {}. Score of {} is {}/{}'.format(self.bat_team.onstrike, self.bat_team.name, self.bat_team.bat_total, self.bat_team.bat_wickets))
            self.bat_team.change_strike()
        
        if res == '4':
            self.bat_team.player_match_stats[self.bat_team.onstrike]['runs_scored'] += 4
            self.bat_team.player_match_stats[self.bat_team.onstrike]['4s'] += 1
            self.bowl_team.player_match_stats[self.bowl_team.bowler]['runs_off'] += 4
            self.bat_team.bat_total += 4
            self.bowl_team.bowl_total += 4
            if self.log_verbosity:
                print('Four runs scored by {}. Score of {} is {}/{}'.format(self.bat_team.onstrike, self.bat_team.name, self.bat_team.bat_total, self.bat_team.bat_wickets))
        
        if res == '6':
            self.bat_team.player_match_stats[self.bat_team.onstrike]['runs_scored'] += 6
            self.bat_team.player_match_stats[self.bat_team.onstrike]['6s'] += 1
            self.bowl_team.player_match_stats[self.bowl_team.bowler]['runs_off'] += 6
            self.bat_team.bat_total += 6
            self.bowl_team.bowl_total += 6
            if self.log_verbosity:
                print('{} hit a six! Score of {} is {}/{}'.format(self.bat_team.onstrike, self.bat_team.name, self.bat_team.bat_total, self.bat_team.bat_wickets))
        
        if res == 'W':
            if self.log_verbosity:
                print('{} got out by {}'.format(self.bat_team.onstrike, self.bowl_team.bowler))
            self.bowl_team.bowl_wickets += 1
            #self.bat_team.player_match_stats[self.bat_team.onstrike]['out'] = 1
            self.bowl_team.player_match_stats[self.bowl_team.bowler]['wickets'] += 1
            self.bat_team.wicket()

        return True
    

    def simulate_over(self):
        self.overs += 1
        commentary = []

        while (self.balls < 6) and (self.bat_team.bat_wickets < 10) and ((self.innings == 1) or (self.bat_team.bat_total <= self.bowl_team.bat_total)):
            bwl = self.bowl_team.bowler
            bat = self.bat_team.onstrike
            res = self.simulate_ball()
            commentary.append([self.innings, self.overs, self.balls, bwl, bat, res])

        self.balls = 0
        self.bat_team.new_over()
        self.bowl_team.new_over()

        return True
    

    def toss(self):
        if np.random.uniform()>0.5:
            self.bat_team, self.bowl_team = self.bowl_team, self.bat_team
        self.bat_team.team_state = 'bat'
        self.bowl_team.team_state = 'bowl'
        if self.log_verbosity:
            print('{} won the toss and choose to bat first against {}.'.format(self.bat_team.name, self.bowl_team.name))
        
        return True
    

    def change_innings(self):
        " Change bat and bowl teams after first innings"
        self.bat_team, self.bowl_team = self.bowl_team, self.bat_team
        self.bat_team.team_state = 'bat'
        self.bowl_team.team_state = 'bowl'
        self.overs = 0
        self.balls = 0
        self.innings = 2

        if self.log_verbosity:
            print('1st innings completed, now {} will bat.'.format(self.bat_team.name))
        
        return True

    
    def simulate_match(self):
        self.toss()
        if self.log_verbosity:
            print('*'*30)
            print('Starting 1st Innings')
            print('*'*30)
        
        while (self.overs < 20) and (self.bat_team.bat_wickets < 10):
            self.simulate_over()
        
        if self.log_verbosity:
            print()
            print('Final Score: {}/{} of {} overs.'.format(self.bat_team.bat_total, self.bat_team.bat_wickets, self.overs))

        self.change_innings()

        if self.log_verbosity:
            print()
            print('*'*30)
            print('Starting second innings')
            print('*'*30)
        #print(self.bat_team.name)
        #print(self.bowl_team.name)

        while (self.overs < 20) and (self.bat_team.bat_wickets < 10) and (self.bat_team.bat_total <= self.bowl_team.bat_total):
                self.simulate_over()

        if self.bat_team.bat_total > self.bowl_team.bat_total:
            winner = self.bat_team.name
        elif self.bat_team.bat_total < self.bowl_team.bat_total:
            winner = self.bowl_team.name
        else:
            winner = 'Tie'

        if self.log_verbosity:
            print()
            print('Final Score: {}/{} of {} overs.'.format(self.bat_team.bat_total, self.bat_team.bat_wickets, self.overs))

            print()
            print('*'*20)
            if winner != 'Tie':
                print('{} WIN!'.format(winner))
            else:
                print('TIE!')
            print('*'*20)
        
        print('Simulation Complete')

        return [winner,
                    [self.bowl_team.name, self.bowl_team.bat_total, self.bowl_team.bat_wickets, self.bowl_team.player_match_stats],
                    [self.bat_team.name, self.bat_team.bat_total, self.bat_team.bat_wickets, self.bat_team.player_match_stats]
                ]

