import pandas as pd
import numpy as np

class Team(object):
    def __init__(self, name, team_players, batsman_stats, bowler_stats):
        self.name = name
        self.team_players = team_players  #List of players, order of the list indicates batting order 
        self.batsman_stats = batsman_stats  #Hisorical batting stats of all players
        self.bowler_stats = bowler_stats  #Historical bowling stats of all players
        
        self.bat_order = {i+1:x for i, x in enumerate(self.team_players)}  # batting order
        self.bowler_probs = {}   #dict holding probability of giving over to bowlers, the more balls they have bowled, the more probability they have
        self.team_state = ''  #Possible values = bat or bowl

        self.player_match_stats = {}   #stats of players for current match
        for player in self.team_players:
            self.player_match_stats[player] =  {
                    'runs_scored': 0,  #bat
                    '4s': 0,            #bat
                    '6s': 0,            #bat
                    'out': 0,           #bat
                    'overs': 0,         #bowl
                    #'maidens': 0,       #bowl
                    'runs_off': 0,      #bowl
                    'wickets': 0        #bowl
                } 

        #Initialize team state variables
        self._init_state()


    def _init_state(self):
        self.bat_total = 0
        self.bat_wickets = 0
        self.bowl_total = 0
        self.bowl_wickets = 0

        self.onstrike = self.bat_order[1]
        self.offstrike = self.bat_order[2]
        
        self.get_player_bowls()
        self.bowler = self.get_next_bowler(first=True)


    def get_next_bowler(self, first=False):
        bowlers = []
        probs = []
        if first:
            self.prev_bowler = ''
        else:
            self.prev_bowler = self.bowler
        
        #Iterate on the players who could be bowlers
        for player, balls in self.bowler_probs.items():
            if player == self.prev_bowler:  #Player should not be the one who completed last over
                #print('cannot select prev_bowler')
                continue
            if self.player_match_stats[player]['overs'] >= 4:
                #print('maximum overs done')
                continue        #Remove players who have completed 4 overs in this match
            bowlers.append(player)
            probs.append(balls)
        
        probs = [x/sum(probs) for x in probs]
        return np.random.choice(a=bowlers, size=1, p=probs)[0]


    def get_player_bowls(self):
        #print('Setting up bowling probabilities')
        for player in self.team_players:
            try:
                player_bowl_stats = self.bowler_stats[player]
                balls_bowled = np.sum(player_bowl_stats)
                if balls_bowled > 30:    #only want to give bowling who have bowled more than 5 overs in lifetime / historically
                    #print('Adding {}'.format(player))
                    self.bowler_probs[player] = balls_bowled
            except:
                print('Player {} not present in historical bowler stats'.format(player))
        #print(self.bowler_probs)
        #total_bowls = np.sum(list(self.bowler_probs.values()))
        ##Converting bowls bowled to probabilities
        #for player,prob in self.bowler_probs.items():
        #    self.bowler_probs[player] = prob/total_bowls
        
        return True


    def get_hist_probs(self, player, type):
        #Returns historical stats of a player
        if type == 'bat':
            if player in self.batsman_stats.keys():
                hist_prob = self.batsman_stats[player]
                hist_prob = hist_prob / np.sum(hist_prob)
            else: 
                print('Warning: {} not found in batsman stats'.format(player))
                hist_prob = [0., 0., 0., 0., 0., 0., 0.]
        
        elif type == 'bowl':
            if player in self.bowler_stats.keys():
                hist_prob = self.bowler_stats[player]
                hist_prob = hist_prob / np.sum(hist_prob)
            else: 
                print('Warning: {} not found in bowler stats'.format(player))
                hist_prob = [0., 0., 0., 0., 0., 0., 0.] 

        else:
            print("Team {}'s state (bat/bowl) not defined".format(self.name))
            raise ValueError("Team {}'s state (bat/bowl) not defined".format(self.name))   

        return hist_prob   


    def change_strike(self):
        # Updates the on-strike batter between overs
        self.onstrike, self.offstrike = self.offstrike, self.onstrike

    def wicket(self):
        # Updates the team after a wicket
        self.bat_wickets += 1  
        self.player_match_stats[self.onstrike]['out'] = 1
        if self.bat_wickets < 10:
            self.onstrike = self.team_players[self.bat_wickets + 1]

    def new_over(self):
        # module to start a new over
        if self.team_state == 'bat':
            self.change_strike()
        elif self.team_state == 'bowl':
            self.bowler = self.get_next_bowler()
            print('New bowler is: {}'.format(self.bowler))
            self.player_match_stats[self.bowler]['overs'] += 1

